
import numpy as np
import torch
from   pykeops.torch import kernel_product, Kernel
from   shapes_toolbox import Curve
import shapes_toolbox as st


# Pseudo-Hausdorff "distances" ==============================================================================


def _distance_field_knn(params, x_i, Beta) :
    """
    Computes an adaptive distance field "b" wrt. the measure "β", 
    and sample it on the point cloud x_i. "b" is such that:
    
                          f(x) = ∫ exp( -C(x,y) / b(x) ) dβ(y)

    is constant on the ambient space, where "C" is a cost function (log-kernel) specified
    by the dict "params".
    As of today, this routines simply iterates an update rule 

            b(x_i) = - b(x_i) * log[ \sum_j exp( -C(x_i,y_j)/b(x_i) ) * β_j ]

    which (kind of) makes sense if you see
                        v_j -> -ε * log sum_j exp (-v_j/ε)
    as a Softmin operator "of strength ε".

    N.B.: This is ongoing work between myself - Jean Feydy - and Alain Trouvé.
    """
    params  = params.copy()

    nits    = params.get("nits", 5) # You should use nits >= 1.

    # Default values for epsilon, which are homogenous to
    #            " C(x_i,y_j) = - log( kilog_M(x_i,y_j) ) "
    # The values below make sense if you use a Gaussian kernel (quadratic cost)
    # on the unit square/cube.
    K       = params.get("k",              .1   )  # Mass fraction
    epsilon = params.get("epsilon",        .1**2)  # First estimate for "u".
    eps_min = params.get("epsilon_min",  .001**2)  # "Point-spread", minimal value - prevents overfit.
    eps_max = params.get("epsilon_max",   100**2)  # Maximum value - prevents numerical overflows.

    beta, y_j = Beta
    beta_sum  = beta.sum()
    beta_norm = beta / beta_sum # N.B.: normalizing Beta is needed for convergence
    beta_log  = beta_norm.log().view(-1,1)

    def threshold(vmin, v, vmax) :
        """
        Iterating pointwise divisions may result in Infs and NaNs,
        especially if "x_i is in the support of Nu"... 
        To prevent this, we make sure that the distance field "v" 
        takes values in "[vmin,vmax]".
        """
        return torch.min( torch.max(vmin, v), vmax )

    X     = x_i[0] if isinstance(x_i, (list,tuple)) else x_i
    N     = X.shape[0]
    B     = epsilon * torch.ones(N,1).type_as(X)
    B_min = eps_min * torch.ones(N,1).type_as(X)
    B_max = eps_max * torch.ones(N,1).type_as(X)
    
    log_K  = float( np.log(K) )
    ilog_K = 1 / log_K

    for i in range(nits) :
        # Update rule :
        # u(x_i) = - u(x_i) * log[ \sum_j exp( -C(x_i,y_j)/u(x_i) ) * nu_j ]
        # This may not converge optimally, but was okay everytime I used it.
        B = ilog_K * B * kernel_product( params, x_i, y_j, beta_log, 1/B, mode="lse_mult_i" ).view(-1,1)
        B = threshold(B_min, B, B_max)
    
    return - log_K * B.view(-1) - beta_sum.log()

def distance_field_knn(params, x_i, Beta) :
    return _distance_field_knn(params, x_i, Beta), _distance_field_knn(params, Beta[1], Beta)



    

def distance_field_sinkhorn(params, x_i, Beta) :
    params  = params.copy()

    nits    = params.get("nits", 1) # Default : -ε*log( k⋆β )
    epsilon = params.get("epsilon", .1**2)
    beta, y_j = Beta
    beta_log  = beta.log().view(-1,1)

    Y = y_j[0] if isinstance(y_j, (list,tuple)) else y_j
    B_y = torch.zeros(Y.shape[0],1).type_as(Y) # B_y = b(y)/eps

    for i in range(nits-1) : # symmetric updates
        B_y = .5*( B_y - kernel_product( params, y_j, y_j, beta_log + B_y, mode="lse" ).view(-1,1) )

    b_x = - epsilon * kernel_product( params, x_i, y_j, beta_log + B_y, mode="lse" ).view(-1)
    b_y = - epsilon * kernel_product( params, y_j, y_j, beta_log + B_y, mode="lse" ).view(-1)

    return b_x, b_y



def integrate( mu, f ) :
    """Computes <μ, f>."""
    return torch.dot( mu.view(-1), f.view(-1) )

def hausdorff(params, Alpha, Beta, info = False) :
    """
    If 
                  b(x) = C(x, β)     and     a(y) = C(α, y)
    are the two adaptive distance fields defined by one of the routines above,
    returns
                        d(α,β) = < α-β , b-a >
    which can be used as a pseudo-Hausdorff distance between the two measures.
    Most importantly (unlike kernel distances), this formula does *not*
    saturate when the two measures get far from each other.
    Its gradient can thus be used prior to any rigid registration (if needed).

    N.B.: If params["nits"] == 1, we retrieve (up to a multiplicative constant)
                   d(α,β) = < α-β , log( k★α / k★β ) >
    where k is a kernel function specified by params["id"] and params["epsilon"].

    Otherwise (say, if nits >= 5), the gradient of "d" gets more spread out
    across the measures, and can be used as a reliable descent
    direction for registration purposes.

                      -----------------------------------

    My personal take on this is that you could use this penalty as a cheap alternative
    to Wasserstein / Optimal Transport distances between measures:
        In order to get a reasonable gradient, you don't have to compute a full
                         point-to-point transport plan!

    Cases where you *need* the *covering guarantee* offered by OT theory
    (Sinkhorn algorithm, etc.) are not too often encountered in practice.
    A typical example would be that of two horizontal bars, sliding next to each other:

                                   *****************************
                                  ************* β ***************
                                   *****************************
          **************************
         ************* α ************
          **************************
    
    In this situation, bar using a fully converged Sinkhorn algorithm (30~100 convolutions), 
    I do not know any method that would allow you to retrieve a nice uniform gradient on μ...
    (less "global" routines, including adaptive Hausdorff, give you
      a stronger gradient on the rightmost part of μ than on the left one.)
    """
    alpha, x_i = Alpha ; beta, y_j = Beta
    distance_name = params.get("distance_field", "sinkhorn")
    formula       = params.get("cost",           "bregman")

    if   distance_name == "knn" :
        distance_field = distance_field_knn
    elif distance_name == "sinkhorn" :
        distance_field = distance_field_sinkhorn

    a_y, a_x = distance_field( params, y_j, Alpha )
    b_x, b_y = distance_field( params, x_i, Beta )

    if formula == "bregman" :
        cost = .5*( integrate( alpha, b_x - a_x  ) \
                  + integrate( beta , a_y - b_y  ) )
    elif formula == "bregman_a" :
        cost = .5* integrate( alpha, b_x - a_x  )
    elif formula == "bregman_b" :
        cost = .5* integrate( beta , a_y - b_y  )
    elif formula == "KL" :
        eps = params.get("epsilon", .1**2)
        cost = .5*( integrate( alpha * (-a_x/eps).exp(), b_x - a_x  ) \
                  + integrate( beta  * (-b_y/eps).exp(), a_y - b_y  ) )


    if cost.item() < 0 :
        print("!")
        #raise ValueError("Interesting !!!")

    heatmaps = None,None
    if info :
        mode = params.get("info", "heatmaps+curves")
        info_1, info_2, a2b, b2a = None, None, None, None

        if mode in ["heatmaps", "heatmaps+curves"] and params.get("features", "locations") == "locations" :

            # Create a uniform grid on the [-2,+2]x[-2,+2] square:
            xmin,xmax,ymin,ymax,res  = params.get("heatmap_range", (0,1,0,1,100))
            ticks_x = np.linspace( xmin, xmax, res + 1)[:-1] + 1/(2*res) 
            ticks_y = np.linspace( ymin, ymax, res + 1)[:-1] + 1/(2*res) 
            X,Y    = np.meshgrid( ticks_x, ticks_y )

            dtype  = alpha.data.type()
            points = torch.from_numpy(np.vstack( (X.ravel(), Y.ravel()) ).T).contiguous().type(dtype)

            # Compute and sample the distance fields "v" and "u" on this grid:
            # N.B.: we use a Sqrt, since it looks better when using a gaussian kernel...

            info_1,_ = distance_field(params, points, Alpha )
            info_2,_ = distance_field(params, points, Beta  )

            info_1 = info_1.view(res,res).data.cpu().numpy()
            info_2 = info_2.view(res,res).data.cpu().numpy()

        if mode in ["curves", "heatmaps+curves"] :
            alpha, beta = alpha.view(-1,1), beta.view(-1,1)
            X = x_i[0] if isinstance(x_i, tuple) else x_i
            Y = y_j[0] if isinstance(y_j, tuple) else y_j
            nx,ny  = len(X),len(Y) 
            X_targets = kernel_product( params, x_i,y_j, Y, 0*alpha, beta.log(),X,  mode="log_scaled_barycenter") \
                      / kernel_product( params, x_i,y_j, beta )

            points = torch.cat( (X, X + X_targets) ).contiguous()
            connec = torch.stack( ( torch.arange( 0,nx ), torch.arange(nx, 2*nx) ) , dim=1).contiguous().type(st.dtypeint)
            a2b  = Curve( points, connec )

            Y_targets = kernel_product( params, y_j,x_i, X, 0*beta, alpha.log(),Y, mode="log_scaled_barycenter") \
                      / kernel_product( params, y_j,x_i, alpha )

            points = torch.cat( (Y, Y + Y_targets) ).contiguous()
            connec = torch.stack( ( torch.arange( 0,ny ), torch.arange(ny, 2*ny) ) , dim=1).contiguous().type(st.dtypeint)
            b2a  = Curve(points, connec)


        # reshape as a pair of "background" images
        heatmaps = ( info_1, info_2, a2b, b2a)

        if formula == "bregman_a" :
            heatmaps = ( None, info_2, a2b, None )
        elif formula == "bregman_b" :
            heatmaps = ( info_1, None, None, b2a )


    return cost, Heatmaps(*heatmaps)



class Heatmaps :
    def __init__(self, a, b, a2b=None, b2a=None) :
        self.a = a
        self.b = b
        self.a2b = a2b
        self.b2a = b2a

    def plot(self, axis, params) :
        par_plot = params.get("display", {})
        coords = params.get("fidelity", {}).get( "heatmap_range", (0,1,0,1,100) )
        def contour_plot(img, color, levels) :
            axis.contour(img, origin='lower', linewidths = 1., colors = color,
                        levels = levels, extent=coords[0:4]) 
            try :
                None
                #axis.contour(img, origin='lower', linewidths = 2., colors = color,
                #            levels = (0.), extent=coords[0:4]) 
            except :
                pass

        if self.a is not None :
            color_a     = par_plot.get("info_color_a",    "#E2C5C5")
            levels_a = np.linspace(np.amin( self.a[:] ), np.amax( self.a[:] ), 30)
            contour_plot(self.a, color_a, levels_a)
            
        if self.b is not None :
            color_b     = par_plot.get("info_color_b",    "#C8DFF9")
            levels_b = np.linspace(np.amin( self.b[:] ), np.amax( self.b[:] ), 30)
            contour_plot(self.b, color_b, levels_b)

        if self.a2b is not None :
            color_a2b = par_plot.get("info_color_a",    (.8, .4, .4, .05))
            lw          = par_plot.get("info_linewidth", 1 )
            self.a2b.plot(axis, color=color_a2b, linewidth=lw)

        if self.b2a is not None :
            color_b2a = par_plot.get("info_color_b",    (.4, .4, .8, .05))
            lw          = par_plot.get("info_linewidth", 1 )
            self.b2a.plot(axis, color=color_b2a, linewidth=lw)

    def save(self, fname) :
        None
        self.a2b.save(fname+"_a2b")
        self.b2a.save(fname+"_b2a")
